import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import datetime
import re
from datetime import date, timedelta
import calendar
from Text2JSON.entity_named_recog.entity_utils import *
from Text2JSON.entity_named_recog.year_entity_recognition import year_absolute_comfirm, year_relative_comfirm


def month_recon(line, placeholders_list: dict):
    line = month_relative_recognition(line, placeholders_list)
    line = month_absolute_recognition(line, placeholders_list)
    return line


# 月相对位置识别
def month_relative_recognition(line, placeholders_list: dict):
    cur_month = ['今个月', '这个月', '本月', '今月', '这月']
    now_month = ['最近一个月', '近一个月', '最近1个月', '近1个月', '近月']
    next_month = ['下1个月', '下一个月', '下一月', '下个月', '下月', '一个月后', '1个月后']
    last_month = ['上1个月', '上一个月', '上一月', '上个月', '上月', '一个月前', '1个月前']
    flag = True
    for month in cur_month:
        if month in line:
            index = line.index(month)
            first_line = line[0: index]
            sec_line = line[index + len(month):]

            # now
            now = date.today()

            # 本月第一天喝最后一天
            this_month_start = datetime.datetime(now.year, now.month, 1)
            this_month_end = datetime.datetime(now.year, now.month, calendar.monthrange(now.year, now.month)[1])

            # to str and insert into line
            time_str = '“' + date_to_str(this_month_start) + "”" + "到" + '“' + date_to_str(this_month_end) + "”"
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [this_month_start, this_month_end])
            line = first_line + placeholder + sec_line

    # 保证最长匹配优先
    while flag:
        flag = False
        for month in now_month:
            if month in line:
                index = line.index(month)
                first_line = line[0: index]
                sec_line = line[index + len(month):]

                # now
                now = date.today()
                # 本月的第一天
                now_month_start = datetime.datetime(now.year, now.month, 1)
                # 上月的最后一天
                last_month_end = now_month_start + timedelta(days=-1)
                cnt = 0
                before_n_days = None
                while True:
                    try:
                        before_n_days = datetime.datetime(last_month_end.year, last_month_end.month, now.day - cnt)
                        break
                    except ValueError:
                        cnt += 1
                        continue
                assert before_n_days is not None, '逻辑错误'
                # to str and insert into line
                time_str = '“' + date_to_str(before_n_days) + "”" + "到" + '“' + date_to_str(now) + "”"
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [before_n_days, now])
                line = first_line + placeholder + sec_line
                flag = True
                break

    for month in next_month:
        if month in line:
            index = line.index(month)
            first_line = line[0: index]
            sec_line = line[index + len(month):]

            now = date.today()
            # 本月最后一天
            this_month_end = datetime.datetime(now.year, now.month, calendar.monthrange(now.year, now.month)[1])

            # 下一月第一天和最后一天
            next_month_start = this_month_end + timedelta(days=1)
            next_month_end = datetime.datetime(next_month_start.year, next_month_start.month,
                                               calendar.monthrange(next_month_start.year,
                                                                   next_month_start.month)[1])
            # to str and insert into line
            time_str = '“' + date_to_str(next_month_start) + '”' + '到' + '“' + date_to_str(next_month_end) + '”'
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [next_month_start, next_month_end])
            line = first_line + placeholder + sec_line

    for month in last_month:
        if month in line:
            index = line.index(month)
            first_line = line[0: index]
            sec_line = line[index + len(month):]

            now = date.today()
            # 当前月第一天
            this_month_start = datetime.datetime(now.year, now.month, 1)
            # 上月最后一天和上个月第一天
            last_month_end = this_month_start - timedelta(days=1)
            last_month_start = datetime.datetime(last_month_end.year, last_month_end.month, 1)

            # to str and insert into line
            time_str = '“' + date_to_str(last_month_start) + '”' + '到' + '“' + date_to_str(last_month_end) + '”'
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [last_month_start, last_month_end])
            line = first_line + placeholder + sec_line

    return line


# 月相对位置识别
def month_absolute_recognition(line, placeholders_list: dict):
    now = date.today()
    month_map = {'一': '1', '二': '2', '三': '3', '四': '4',
                 '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}
    # 数字 月份 10,11,12
    date_all = re.findall(r"(\d{1,2}月[份]?)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        month = int(re.sub(r'(月[份]?)', '', data))

        handle_line, year = year_absolute_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year = year_relative_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year

        try:
            month_start = datetime.datetime(year, month, 1)
            month_end = datetime.datetime(year, month,
                                          calendar.monthrange(year, month)[1])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None, None])
            line = first_line + placeholder + sec_line
            continue
        # 用占位符暂时代替，否则就会无限重复
        time_str = '“' + date_to_str(month_start) + "”" + "到" + '“' + date_to_str(month_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [month_start, month_end])
        line = handle_line + placeholder + sec_line

    date_all = re.findall(r"([第]?十[一二]?[个]?月[份]?)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        month = re.sub(r'([个]?月[份]?)', '', data)
        month = re.sub(r'([第])', '', month)

        if len(month) == 1:
            month = 10
        else:
            month = 10 + int(month_map[re.sub(r'(十)', '', data)])

        handle_line, year = year_absolute_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year = year_relative_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year

        try:
            month_start = datetime.datetime(year, month, 1)
            month_end = datetime.datetime(year, month,
                                          calendar.monthrange(year, month)[1])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None, None])
            line = first_line + placeholder + sec_line
            continue
        # 用占位符暂时代替，否则就会无限重复
        time_str = '“' + date_to_str(month_start) + "”" + "到" + '“' + date_to_str(month_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [month_start, month_end])
        line = handle_line + placeholder + sec_line

    date_all = re.findall(r"([第]?[一二三四五六七八九][个]?月[份]?)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        month = re.sub(r'([个]?月[份]?)', '', data)
        month = re.sub(r'([第])', '', month)
        month = int(month_map[month])
        handle_line, year = year_absolute_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year = year_relative_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year

        try:
            month_start = datetime.datetime(year, month, 1)
            month_end = datetime.datetime(year, month,
                                          calendar.monthrange(year, month)[1])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None, None])
            line = first_line + placeholder + sec_line
            continue
        # 用占位符暂时代替，否则就会无限重复
        time_str = '“' + date_to_str(month_start) + "”" + "到" + '“' + date_to_str(month_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [month_start, month_end])
        line = handle_line + placeholder + sec_line

    date_all = re.findall(r"(最后[一1][个]?月[份]?)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        month = 12
        handle_line, year = year_absolute_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year = year_relative_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year

        try:
            month_start = datetime.datetime(year, month, 1)
            month_end = datetime.datetime(year, month,
                                          calendar.monthrange(year, month)[1])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None, None])
            line = first_line + placeholder + sec_line
            continue
        # 用占位符暂时代替，否则就会无限重复
        time_str = '“' + date_to_str(month_start) + "”" + "到" + '“' + date_to_str(month_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [month_start, month_end])
        line = handle_line + placeholder + sec_line
    return line


def month_absolute_comfirm(line, placeholders_list: dict):
    # 需要找到最后出现的位置, '....2013年...' 不识别, '...2014年'识别
    # 重复出现,"2014年...,2014年"只识别最后一个
    month = None
    year = None
    now = date.today()
    month_map = {'一': '1', '二': '2', '三': '3', '四': '4',
                 '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}
    connect_char = ['-', '~', '到', '至']

    date_all = re.findall(r"(\d{1,2}月[份]?)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            month = int(data.replace('月', ''))
            line = line[:index]

            line, year = year_absolute_comfirm(line, placeholders_list)
            if year is None:
                line, year = year_relative_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            return line, year, month

    # 处理10月,11月,12月
    date_all = re.findall(r'([十][一二]?月[份]?)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            # 转成中文数字
            data = data.replace('月', '')
            data = data.replace('十', '')
            month_str = '0'
            for d_char in data:
                month_str += month_map[d_char]
            month = 10 + int(month_str)
            line, year = year_absolute_comfirm(line, placeholders_list)
            if year is None:
                line, year = year_relative_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            return line, year, month

    # 处理1-9月
    date_all = re.findall(r"([一二三四五六七八九]月[份]?)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            # 转成中文数字
            data = data.replace('月', '')
            month_str = ''
            for d_char in data:
                month_str += month_map[d_char]
            month = int(month_str)
            line = line[:index]
            line, year = year_absolute_comfirm(line, placeholders_list)
            if year is None:
                line, year = year_relative_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            return line, year, month

    if len(line) >= 16 and line[-1] in connect_char:
        holder = line[-16:-1]
        if holder in placeholders_list.keys():
            _, time_list = placeholders_list[holder]
            if len(time_list) == 1 and time_list[0] is not None:
                return line[:-1] + '到', time_list[0].year, time_list[0].month

    return line, year, month


# 月相对位置确认
def month_relative_comfirm(line, placeholders_list: dict):
    # 只识别最后一个词
    cur_month = ['今月', '本月', '这个月', '这月']
    next_month = ['下1个月', '下一个月', '下一月', '下个月', '下月', '一个月后', '1个月后']
    last_month = ['上1个月', '上一个月', '上一月', '上个月', '上月', '一个月前', '1个月前']
    year = None
    month = None
    # 需要把年加进来
    now = date.today()
    for data in cur_month:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                year = now.year
                month = now.month
                line = line[:index]
                return line, year, month
    for data in next_month:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                month_end = datetime.datetime(now.year, now.month, calendar.monthrange(now.year, now.month)[1])
                next_month_start = month_end + timedelta(days=1)
                year = next_month_start.year
                month = next_month_start.month
                line = line[:index]
                return line, year, month

    for data in last_month:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                month_start = datetime.datetime(now.year, now.month, 1)
                last_month_end = month_start - timedelta(days=1)
                year = last_month_end.year
                month = last_month_end.month
                line = line[:index]
                return line, year, month
    return line, year, month


if __name__ == '__main__':
    ''
    # placeholders = {}
    # example1 = '明年11月, 今年12月'
    # sentence = month_recon(example1, placeholders)
    # for holder, data in placeholders.items():
    #     sentence = sentence.replace(holder, data[0])
    # print(example1)
    # print('10', sentence)
    # example1 = '最近一个月， 最近一个月'
    # placeholders_list = []
    # sentence, placeholders = month_absolute_recon(example1)
    # placeholders_list.extend(placeholders)
    # sentence, placeholders = month_relative_recon(sentence)
    # placeholders_list.extend(placeholders)
    # for token, holder in placeholders_list:
    #     sentence = sentence.replace(holder, token)
    # print(sentence)
    #
    # example2 = '去年2月，今年6月份，下年7月，16年6月，2017年十二月，16年二月, 一六年十一月，二零零零年十二月'
    # placeholders_list = []
    # sentence, placeholders = month_absolute_recon(example2)
    # placeholders_list.extend(placeholders)
    # sentence, placeholders = month_relative_recon(sentence)
    # placeholders_list.extend(placeholders)
    # for token, holder in placeholders_list:
    #     sentence = sentence.replace(holder, token)
    # print(sentence)
    #
    # example3 = 'sss12月, 去年12月'
    # example4 = 'sss十二月, 今年二月'
    # sentence, year, month = month_absolute_comfirm(example3)
    # print(sentence, year, month)
    # sentence, year, month = month_absolute_comfirm(example4)
    # print(sentence, year, month)
    # example4 = 'sss十二月, 明年四月'
    # sentence, year, month = month_absolute_comfirm(example4)
    # print(sentence, year, month)
    # example4 = 'sss十二月, 19年四月'
    # sentence, year, month = month_absolute_comfirm(example4)
    # print(sentence, year, month)
    # example4 = 'sss十二月, 零零年四月'
    # sentence, year, month = month_absolute_comfirm(example4)
    # print(sentence, year, month)
    #
    # example5 = 'sss下一个月'
    # sentence, year, month = month_relative_comfirm(example5)
    # print(sentence, year, month)
    #
    # example5 = 'sss上一个月'
    # sentence, year, month = month_relative_comfirm(example5)
    # print(sentence, year, month)



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import numpy as np
import torch.utils.data as torch_data
from Text2JSON.featurizer.schema_parser import schema_parsed
from Text2JSON.featurizer.Input_SQL import SQLExample
from collections import defaultdict

stats = defaultdict(int)


class SchemaInputFeature(object):
    def __init__(self, qid, question, schema_id, relative_schema_id):
        self.qid = qid
        self.question = question
        self.schema_id = schema_id
        self.relative_schema_id = relative_schema_id
        self.tokens = []
        # 用于训练和测试
        self.input_ids = []
        self.input_mask = []
        self.segment_ids = []
        self.schema_rel = []
        self.schema_mask = []

        # 用于测试
        self.schema_sec_rel = []

    def to_dict(self):
        return {
            'qid': self.qid,
            'question': self.question,
            'schema_id': self.schema_id,
            'relative_schema_id': self.relative_schema_id,
            'tokens': self.tokens,
            'input_ids': self.input_ids,
            'input_mask': self.input_mask,
            'segment_ids': self.segment_ids,
            'table_rel': self.schema_rel,
            'mask': self.schema_mask
        }


class HydraSchemaFeaturizer(object):
    """
        特征提取器，将一条训练数据集，转变成为InputFeature
    """

    def __init__(self, config, tokenizer):
        self.config = config
        # 根据不同模型，设置tokenizer
        self.tokenizer = tokenizer

    def get_input_schema_feature(self, example: SQLExample, schema: schema_parsed, config, positive=True, negative=True,
                                note=False):
        max_query_length = int(config["max_query_length"])
        max_column_length = int(config["max_column_length"])
        max_total_length = max_column_length + max_query_length

        # 一条feature包含一条train,包含多个<>column<>question<>编码
        input_feature = SchemaInputFeature(
            example.qid,
            example.question,
            example.table_id,
            example.table_id_relative,
        )

        # 记录映射关系
        tokens = []
        word_to_subword_base = []  # 词 到 子词 的映射关系,形式(start_index, end_index)
        subword_to_word_base = []  # 子词 到 词 的映射关系,形式(0,1,1,2,2,2)
        for i, query_token in enumerate(example.tokens):
            sub_tokens = self.tokenizer.tokenize(query_token)
            cur_pos = len(tokens)
            if len(sub_tokens) > 0:
                word_to_subword_base += [(cur_pos, cur_pos + len(sub_tokens))]
                tokens.extend(sub_tokens)
                subword_to_word_base.extend([i] * len(sub_tokens))

        # positive 只加载正例， negative 只加载负例， 两个都为true，一起加载
        # 遍历表中所有columns，为每一个column生成<>column<>question<>编码
        # 这里只要是表相关的都作为正例
        rel_table = [example.table_id]
        # 是否需要对输入进行拓展
        if note:
            feature_type = 'feature_schema'

        else:
            feature_type = 'feature_schema_without_note'

        for key, value in schema.extend_schema.items():
            # 只要正例
            if positive and not negative:
                if key not in rel_table:
                    continue
            # 只要负例
            elif not positive and negative:
                if key in rel_table:
                    continue

            # 正例和负例都要
            elif positive and negative:
                ''

            table = schema.extend_schema[key]
            for table_column_feature in table[feature_type]:
                # 编码，输入为type, table, column, table_note, column_note, space, token
                # 输出[cls]type, table, column, table_note, column_note, space[SEP]question[SEP]
                assert len(
                    table_column_feature) <= max_column_length, "{0} URL中参数特征长度过大，请缩小参数参数特征或将配置文件中的" \
                                                                "max_column_length增大".format(table['url'])
                assert len(''.join(
                    tokens)) <= max_query_length, '问句: {0} 输入问句长度大于最大字符长度，请将配置文件中的max_query_length' \
                                                  '增大或者缩短输入问句长度'.format(example.question)
                tokenize_result = self.tokenizer.encode_plus(
                    table_column_feature,
                    tokens,
                    max_length=max_total_length,
                    truncation_strategy="longest_first",
                    pad_to_max_length=True,
                )

                input_ids = tokenize_result["input_ids"]  # 将输入tokens, 通过词表转换成id
                segment_ids = tokenize_result["token_type_ids"]  # 标识column和question位置[0,0,1,1]
                input_mask = tokenize_result["attention_mask"]  # 标识column+question位置编码,也就是整个输入所占的编码
                new_tokens = self.tokenizer.convert_ids_to_tokens(input_ids)  # 获取编码后的序列[cls]column[SEP]question[SEP]

                assert len(input_ids) == max_total_length
                assert len(input_mask) == max_total_length
                assert len(segment_ids) == max_total_length

                input_feature.tokens.append(new_tokens)
                input_feature.input_ids.append(input_ids)
                input_feature.input_mask.append(input_mask)
                input_feature.segment_ids.append(segment_ids)

        return input_feature

    def get_predict_input_schema_feature(self, example: SQLExample, schema: schema_parsed, config, note=True):
        max_query_length = int(config["max_query_length"])
        max_column_length = int(config["max_column_length"])
        max_total_length = max_column_length + max_query_length

        # 一条feature包含一条train,包含多个<>column<>question<>编码
        input_feature = SchemaInputFeature(
            example.qid,
            example.question,
            example.table_id,
            example.table_id_relative
        )

        # 记录映射关系
        tokens = []
        word_to_subword_base = []  # 词 到 子词 的映射关系,形式(start_index, end_index)
        subword_to_word_base = []  # 子词 到 词 的映射关系,形式(0,1,1,2,2,2)
        for i, query_token in enumerate(example.tokens):
            sub_tokens = self.tokenizer.tokenize(query_token)
            cur_pos = len(tokens)
            if len(sub_tokens) > 0:
                word_to_subword_base += [(cur_pos, cur_pos + len(sub_tokens))]
                tokens.extend(sub_tokens)
                subword_to_word_base.extend([i] * len(sub_tokens))

        # 是否需要对输入进行拓展
        if note:
            feature_type = 'feature_schema'

        else:
            feature_type = 'feature_schema_without_note'

        for key, value in schema.extend_schema.items():
            table = schema.extend_schema[key]
            for table_column_feature in table[feature_type]:
                # 编码，输入为type, table, column, table_note, column_note, space, token
                # 输出[cls]type, table, column, table_note, column_note, space[SEP]question[SEP]
                assert len(table_column_feature) <= max_column_length, "列特征长度过小，请将配置文件中的max_column_length增大"
                assert len(''.join(tokens)) <= max_query_length, "输入问句长度大于最大字符长度，请将配置文件中的max_query_length" \
                                                                 "增大或者缩短输入问句长度"
                tokenize_result = self.tokenizer.encode_plus(
                    table_column_feature,
                    tokens,
                    max_length=max_total_length,
                    truncation_strategy="longest_first",
                    pad_to_max_length=True,
                )

                input_ids = tokenize_result["input_ids"]  # 将输入tokens, 通过词表转换成id
                segment_ids = tokenize_result["token_type_ids"]  # 标识column和question位置[0,0,1,1]
                input_mask = tokenize_result["attention_mask"]  # 标识column+question位置编码,也就是整个输入所占的编码
                new_tokens = self.tokenizer.convert_ids_to_tokens(input_ids)  # 获取编码后的序列[cls]column[SEP]question[SEP]

                assert len(input_ids) == max_total_length, 'input_ids, 错误'
                assert len(input_mask) == max_total_length, 'input_mask, 错误'
                assert len(segment_ids) == max_total_length, 'segment_ids, 错误'

                input_feature.tokens.append(new_tokens)
                input_feature.input_ids.append(input_ids)
                input_feature.input_mask.append(input_mask)
                input_feature.segment_ids.append(segment_ids)

        return input_feature

    def fill_label_schema_feature_split(self, example: SQLExample, input_feature: SchemaInputFeature,
                                       schema: schema_parsed, positive=True, negative=True):
        # 只要正例,对应训练
        if positive and not negative:
            global_table_id = 0
        # 只要负例，对应训练
        elif not positive and negative:
            global_table_id = None
        # 正例和负例都要,对应测试
        elif positive and negative:
            global_table_id = schema.table_local_to_global[str(example.table_id)]
        else:
            assert False, 'Positive 和 Negative 不能全为负'

        # set schema number
        schema_num = len(input_feature.input_ids)  # 表的数量

        # set schema relative
        input_feature.schema_rel = [0] * schema_num
        if global_table_id is not None:
            input_feature.schema_rel[global_table_id] = 1

        # set mask
        input_feature.schema_mask = [example.seq] * schema_num

        return True

    # 加载数据
    def load_schema_data(self, data_paths, config, schema, include_label=False, positive=True, negative=True,
                         note=False):
        '''
            data_paths: 数据路径
            config: 模型参数设置
            include_label: 是否包含标记label
        '''
        # 将input,封装到model_input 中
        # 每一条数据是<>col<>question<>的特征
        model_inputs = {k: [] for k in ["input_ids", "input_mask", "segment_ids"]}
        # 拓展
        if include_label:
            for k in ['schema_rel', 'schema_mask']:
                model_inputs[k] = []

        # 记录输入每一个question所对应的特征序列
        pos = []
        # 多条<>column<>question<>特征
        input_features = []

        # 支持多个数据集
        for data_path in data_paths.split("|"):
            cnt = 0
            for line in open(data_path, encoding="utf8"):
                # 加载一个样本, 一行里面包含一个问句，一个问句有可能对应两个SQL，按SQL进行拆分，拆分成两个样本
                # 将所有正负样本标记在一起
                for index, example in enumerate(SQLExample.load_from_json(line)):
                    # 丢弃未经过验证并且包含sql_label的数据
                    if not example.valid and include_label == True:
                        continue

                    # 获取输入特征,返回一个对象，里面封装了多个<col,question>,<col,question>
                    input_feature = self.get_input_schema_feature(example, schema, config, positive, negative, note)

                    # 如果包含sql_label, 则需要填充特征
                    if include_label:
                        success = self.fill_label_schema_feature_split(example, input_feature, schema, positive,
                                                                       negative)
                        if not success:
                            continue

                    input_features.append(input_feature)  # 总特征

                    # 当前模型有多少个特征
                    cur_start = len(model_inputs["input_ids"])
                    # 当前样本输入有多少个输入特征
                    cur_sample_num = len(input_feature.input_ids)
                    # 标记当前输入模型对应特征的开始和结束位置
                    pos.append((cur_start, cur_start + cur_sample_num))

                    model_inputs["input_ids"].extend(input_feature.input_ids)
                    model_inputs["input_mask"].extend(input_feature.input_mask)
                    model_inputs["segment_ids"].extend(input_feature.segment_ids)
                    if include_label:
                        model_inputs['schema_rel'].extend(input_feature.schema_rel)
                        model_inputs['schema_mask'].extend(input_feature.schema_mask)
                    else:
                        # 如果 include_label is False，则说明这是预测时用的数据，那么只需要一条特征就行了
                        # 因为可以对一条特征进行两次预测，分别是SQL1 和 SQL2
                        break
                    cnt += 1
                    if cnt % 5000 == 0:
                        print('featuring:', cnt)


        '''
            input_features : 数据的特征，可以认为是label
            model_inputs : 数据输入
            pos :  一条数据与多条<>column<>question<>的对应关系
        '''
        return input_features, model_inputs, pos

    # 加载数据
    def load_schema_predict_data(self, example: SQLExample, config, schema, note=True):
        '''
            data_paths: 数据路径
            config: 模型参数设置
            include_label: 是否包含标记label
        '''
        # 将input,封装到model_input 中
        # 每一条数据是<>col<>question<>的特征
        model_inputs = {k: [] for k in ["input_ids", "input_mask", "segment_ids"]}

        # 记录输入每一个question所对应的特征序列
        pos = []
        # 多条<>column<>question<>特征
        input_features = []

        # 获取输入特征,返回一个对象，里面封装了多个<col,question>,<col,question>
        input_feature = self.get_predict_input_schema_feature(example, schema, config, note)

        input_features.append(input_feature)  # 总特征

        # 当前模型有多少个特征
        cur_start = len(model_inputs["input_ids"])
        # 当前样本输入有多少个输入特征
        cur_sample_num = len(input_feature.input_ids)
        # 标记当前输入模型对应特征的开始和结束位置
        pos.append((cur_start, cur_start + cur_sample_num))

        model_inputs["input_ids"].extend(input_feature.input_ids)
        model_inputs["input_mask"].extend(input_feature.input_mask)
        model_inputs["segment_ids"].extend(input_feature.segment_ids)

        '''
            input_features : 数据的特征，可以认为是label
            model_inputs : 数据输入
            pos :  一条数据与多条<>column<>question<>的对应关系
        '''
        return input_features, model_inputs, pos


class SchemaDataset(torch_data.Dataset):
    def __init__(self,  config, featurizer: HydraSchemaFeaturizer, schema):
        self.config = config
        self.featurizer = featurizer
        self.schema = schema
        self.input_features = None
        self.model_inputs = None
        self.pos = None

    def loads(self, data_paths, include_label=False, positive=True, negative=True, note=False):
        self.input_features, self.model_inputs, self.pos = self.featurizer.load_schema_data(data_paths,
                                                                                            self.config,
                                                                                            self.schema,
                                                                                            include_label,
                                                                                            positive,
                                                                                            negative,
                                                                                            note)
        self.trans_data_type()
        print("file {0} loaded".format(data_paths))

    def load_for_predict(self, example, note=True):
        self.input_features, self.model_inputs, self.pos = self.featurizer.load_schema_predict_data(example,
                                                                                                    self.config,
                                                                                                    self.schema,
                                                                                                    note)
        self.trans_data_type()

    # 转换数据类型
    def trans_data_type(self):
        for k in self.model_inputs:
            self.model_inputs[k] = np.array(self.model_inputs[k], dtype=np.int64)

    # 这两个函数是dataloader的关键
    def __len__(self):
        return self.model_inputs["input_ids"].shape[0]

    def __getitem__(self, idx):
        return {k: v[idx] for k, v in self.model_inputs.items()}


if __name__ == "__main__":
    ''



import sys
from os.path import dirname, abspath

path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.featurizer.column_featurizer import ColumnDataset, HydraColumnFeaturizer
from Text2JSON.featurizer.schema_featurizer import SchemaDataset, HydraSchemaFeaturizer


def load_train_column_data(config, schema, featurizer: HydraColumnFeaturizer, include_label=True, positive=True,
                           negative=True, note=False):
    train_data = ColumnDataset(config, featurizer, schema)
    train_data.loads(config["train_data_path"], include_label, positive, negative, note)
    return train_data


def load_eval_column_data(config, schema, featurizer: HydraColumnFeaturizer, include_label=True, positive=True,
                          negative=True, note=False):
    eval_path = config["dev_data_path"]
    data = ColumnDataset(config, featurizer, schema)
    data.loads(eval_path, include_label, positive, negative, note)
    print("Eval Data file {0} loaded".format(eval_path))
    return data


def load_train_schema_data(config, schema, featurizer: HydraSchemaFeaturizer, include_label=True, positive=True,
                           negative=True, note=False):
    train_schema_data = SchemaDataset(config, featurizer, schema)
    train_schema_data.loads(config["train_data_path"], include_label, positive, negative, note)
    print("Train Data file {0} loaded".format(config["train_data_path"]))
    return train_schema_data


def load_dev_schema_data(config, schema, featurizer: HydraSchemaFeaturizer, include_label=True, positive=True,
                         negative=True, note=False):
    # 这里需要重新加载
    eval_path = config["dev_data_path"]
    data = SchemaDataset(config, featurizer, schema)
    data.loads(eval_path, include_label, positive, negative, note)
    print("Dev Data file {0} loaded".format(config["dev_data_path"]))
    return data



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
from Text2JSON.entity_named_recog.entity_utils import *


def proper_recon(line, placeholders_list: dict):
    char_span = ''
    flag = False
    for char in line:
        if char == '“':
            char_span = char
            flag = True
            continue
        elif flag and char == '”':
            flag = False
            char_span += char
            index = line.index(char_span)
            first_line = line[0: index]
            sec_line = line[index + len(char_span):]
            placeholder = random_str()
            placeholders_list[placeholder] = (char_span, [None])
            line = first_line + placeholder + sec_line
        elif flag:
            char_span += char
    return line

# if __name__ == '__main__':
#     line = '我想看一下“xx民事纠纷“的“二审”判决书'
#     placeholders_list = {}
#     line = proper_noun_recon(line, placeholders_list)
#     for holder, data in placeholders_list.items():
#         line = line.replace(holder, data[0])
#     print(line)
#     print(placeholders_list)


import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import datetime
import re
from datetime import date, timedelta
from Text2JSON.entity_named_recog.day_entity_recognition import day_absolute_comfirm, day_relative_comfirm
from Text2JSON.entity_named_recog.entity_utils import *


def hour_recon(line, placeholders_list: dict):
    line = hour_relative_recognition(line, placeholders_list)
    line = hour_absolute_recognition(line, placeholders_list)
    return line


def hour_absolute_recognition(line, placeholders_list: dict):
    time_map = {'零': '0', '一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '十': '10', '两': '2'}
    now = date.today()
    # 数字 上午
    date_all_morning_1 = re.findall(r"(上午\d{1,2}点)", line)
    date_all_morning_2 = re.findall(r"(早上\d{1,2}点)", line)
    for data in date_all_morning_1 + date_all_morning_2:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(上午)', '', data)
        hour_minute = re.sub(r'(早上)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')

        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(hour_minute[:hour_minute.index(':')])
        try:
            # 进一天
            if hour == 24:
                hour = 0
                target_time = datetime.datetime(year, month, day, hour)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字 下午, 傍晚，晚上
    date_all_afternoon = re.findall(r"(下午\d{1,2}点)", line)
    date_all_nightfall = re.findall(r"(傍晚\d{1,2}点)", line)
    date_all_night = re.findall(r"(晚上\d{1,2}点)", line)
    for data in date_all_afternoon + date_all_nightfall + date_all_night:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(下午)', '', data)
        hour_minute = re.sub(r'(傍晚)', '', hour_minute)
        hour_minute = re.sub(r'(晚上)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(hour_minute[:hour_minute.index(':')])
        try:
            # 进一天
            if hour == 24:
                hour = 0
                target_time = datetime.datetime(year, month, day, hour)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
                # 加时间
                if hour <= 12:
                    target_time = target_time + timedelta(hours=12)

        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字 中午
    date_all_1 = re.findall(r"(中午[01]?[123]点)", line)
    date_all_2 = re.findall(r"(正午[01]?[123]点)", line)
    for data in date_all_1 + date_all_2:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(中午)', '', data)
        hour_minute = re.sub(r'(正午)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(hour_minute[:hour_minute.index(':')])
        try:
            # 进一天
            if hour == 24:
                hour = 0
                target_time = datetime.datetime(year, month, day, hour)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
                # 加时间
                if hour <= 2:
                    target_time = target_time + timedelta(hours=12)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字 凌晨
    date_all_early = re.findall(r"(凌晨[0]?[123456]点)", line)
    for data in date_all_early:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(凌晨)', '', data)
        hour_minute = hour_minute.replace('点', ':')
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(hour_minute[:hour_minute.index(':')])
        try:
            # 进一天
            if hour == 24:
                hour = 0
                target_time = datetime.datetime(year, month, day, hour)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字
    date_all = re.findall(r"(\d{1,2}点)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        hour_minute = data.replace('点', ':')

        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(hour_minute[:hour_minute.index(':')])
        try:
            # 进一天
            if hour == 24:
                hour = 0
                target_time = datetime.datetime(year, month, day, hour)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 中文 上午
    date_all_morning_1 = re.findall(r"(上午[二]?[十]?[零两一二三四五六七八九十]点)", line)
    date_all_morning_2 = re.findall(r"(早上[二]?[十]?[零两一二三四五六七八九十]点)", line)
    for data in date_all_morning_1 + date_all_morning_2:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(上午)', '', data)
        hour_minute = re.sub(r'(早上)', '', hour_minute)
        hour_minute = re.sub(r'(分)', '', hour_minute)
        hour_minute = re.sub(r'(钟)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')

        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = hour_minute[:hour_minute.index(':')]

        # 解析小时
        if len(hour) > 1:
            if '十' not in hour:
                hour = 25
            else:
                pre_hour = hour[0:hour.index('十')]
                post_hour = hour[hour.index('十') + 1:]
                # 10
                if len(pre_hour) == 0 and len(post_hour) == 0:
                    hour = 10
                # 11 ~ 19
                elif len(pre_hour) == 0 and len(post_hour) != 0:
                    hour = 10 + int(time_map[post_hour])
                # 20, 30
                elif len(pre_hour) != 0 and len(post_hour) == 0:
                    hour = int(time_map[pre_hour]) * 10
                else:
                    # 21 ~ 29
                    hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
        else:
            hour = int(time_map[hour])

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 中文 下午
    date_all_afternoon = re.findall(r"(下午[二]?[十]?[两一二三四五六七八九十]点)", line)
    date_all_nightfall = re.findall(r"(傍晚[二]?[十]?[两一二三四五六七八九十]点)", line)
    date_all_night = re.findall(r"(晚上[二]?[十]?[两一二三四五六七八九十]点)", line)
    for data in date_all_afternoon + date_all_nightfall + date_all_night:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(下午)', '', data)
        hour_minute = re.sub(r'(傍晚)', '', hour_minute)
        hour_minute = re.sub(r'(晚上)', '', hour_minute)
        hour_minute = re.sub(r'(分)', '', hour_minute)
        hour_minute = re.sub(r'(钟)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')

        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = hour_minute[:hour_minute.index(':')]

        # 解析小时
        if len(hour) > 1:
            if '十' not in hour:
                hour = 25
            else:
                pre_hour = hour[0:hour.index('十')]
                post_hour = hour[hour.index('十') + 1:]
                # 10
                if len(pre_hour) == 0 and len(post_hour) == 0:
                    hour = 10
                # 11 ~ 19
                elif len(pre_hour) == 0 and len(post_hour) != 0:
                    hour = 10 + int(time_map[post_hour])
                # 20, 30
                elif len(pre_hour) != 0 and len(post_hour) == 0:
                    hour = int(time_map[pre_hour]) * 10
                else:
                    # 21 ~ 29
                    hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
        else:
            hour = int(time_map[hour])

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
                # 加时间
                if hour <= 12:
                    target_time = target_time + timedelta(hours=12)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 中文 中午
    date_all_1 = re.findall(r"(中午十[一二三]点)", line)
    date_all_2 = re.findall(r"(正午十[一二三]点)", line)
    for data in date_all_1 + date_all_2:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(中午)', '', data)
        hour_minute = re.sub(r'(正午)', '', hour_minute)
        hour_minute = re.sub(r'(分)', '', hour_minute)
        hour_minute = re.sub(r'(钟)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = hour_minute[:hour_minute.index(':')]

        # 解析小时
        if len(hour) > 1:
            if '十' not in hour:
                hour = 25
            else:
                pre_hour = hour[0:hour.index('十')]
                post_hour = hour[hour.index('十') + 1:]
                # 10
                if len(pre_hour) == 0 and len(post_hour) == 0:
                    hour = 10
                # 11 ~ 19
                elif len(pre_hour) == 0 and len(post_hour) != 0:
                    hour = 10 + int(time_map[post_hour])
                # 20, 30
                elif len(pre_hour) != 0 and len(post_hour) == 0:
                    hour = int(time_map[pre_hour]) * 10
                else:
                    # 21 ~ 29
                    hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
        else:
            hour = int(time_map[hour])

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)

        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 中文 中午
    date_all_1 = re.findall(r"(中午一点)", line)
    date_all_2 = re.findall(r"(正午一点)", line)
    for data in date_all_1 + date_all_2:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        hour = 13

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 中文 凌晨
    date_all_early = re.findall(r"(凌晨[二]?[十]?[一两二三四五六七八九十]点)", line)
    for data in date_all_early:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        hour_minute = re.sub(r'(凌晨)', '', data)
        hour_minute = re.sub(r'(分)', '', hour_minute)
        hour_minute = re.sub(r'(钟)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')
        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = int(time_map[hour_minute[:hour_minute.index(':')]])

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line

    # 数字
    date_all = re.findall(r"([二]?[十]?[两一二三四五六七八九十]点)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        hour_minute = re.sub(r'(分)', '', data)
        hour_minute = re.sub(r'(钟)', '', hour_minute)
        hour_minute = hour_minute.replace('点', ':')

        # 确定年月日
        handle_line, year, month, day = day_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month, day = day_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        day = now.day if day is None else day

        # 确定小时 和 分钟
        hour = hour_minute[:hour_minute.index(':')]

        # 解析小时
        if len(hour) > 1:
            if '十' not in hour:
                hour = 25
            else:
                pre_hour = hour[0:hour.index('十')]
                post_hour = hour[hour.index('十') + 1:]
                # 10
                if len(pre_hour) == 0 and len(post_hour) == 0:
                    hour = 10
                # 11 ~ 19
                elif len(pre_hour) == 0 and len(post_hour) != 0:
                    hour = 10 + int(time_map[post_hour])
                # 20, 30
                elif len(pre_hour) != 0 and len(post_hour) == 0:
                    hour = int(time_map[pre_hour]) * 10
                else:
                    # 21 ~ 29
                    hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
        else:
            hour = int(time_map[hour])

        try:
            # 进一天
            if hour == 24:
                target_time = datetime.datetime(year, month, day, 0)
                target_time = target_time + timedelta(days=1)
            else:
                target_time = datetime.datetime(year, month, day, hour)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_time is not None, '逻辑错误'
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = handle_line + placeholder + sec_line
    return line


def hour_relative_recognition(line, placeholders_list: dict):
    time_map = {'一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '两': '2'}
    # 半单独拿出来，放到分钟那里
    now = datetime.datetime.now()

    # 中文前 0-99
    date_all = re.findall(r"([二三四五六七八九]?[十]?[零一二三四五六七八九十半][个]?小时前)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 计算时差，获取目标时间
        hour = re.sub(r"[个]?小时前", '', data)
        try:
            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    raise ValueError
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        target_time = now - timedelta(hours=hour)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 数字前
    date_all = re.findall(r"(\d{1,2}[个]?小时前)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        data = re.sub(r"[个]?小时前", '', data)
        hours_delta = int(data)
        target_time = now - timedelta(hours=hours_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 中文后 10-99
    date_all = re.findall(r"([二三四五六七八九]?[十]?[零一二三四五六七八九十半][个]?小时后)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 计算时差，获取目标时间
        data = re.sub(r"[个]?小时后", '', data)
        hour = data
        try:
            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    raise ValueError
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        target_time = now + timedelta(hours=hour)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 数字后
    date_all = re.findall(r"(\d{1,2}[个]?小时后)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        data = re.sub(r"[个]?小时后", '', data)
        hours_delta = int(data)
        target_time = now + timedelta(hours=hours_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 半 前
    date_all = re.findall(r"(半[个]?小时前)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        minute_delta = 30
        target_time = now - timedelta(minutes=minute_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line

    # 半 后
    date_all = re.findall(r"(半[个]?小时后)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]
        # 计算时差，获取目标时间
        minute_delta = 30
        target_time = now + timedelta(minutes=minute_delta)
        time_str = '“' + time_to_str(target_time) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_time])
        line = first_line + placeholder + sec_line
    return line


def hour_absolute_comfirm(line, placeholders_list: dict):
    time_map = {'零': '0', '一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '十': '10', '两': '2'}
    connect_char = ['-', '~', '到', '至']
    now = date.today()
    # 数字 上午
    date_all_morning_1 = re.findall(r"(上午\d{1,2}[点:])", line)
    date_all_morning_2 = re.findall(r"(早上\d{1,2}[点:])", line)
    for data in date_all_morning_1 + date_all_morning_2:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(上午)', '', data)
            hour = re.sub(r'(早上)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day
            hour = int(hour)
            return line, year, month, day, hour

    # 数字 下午, 傍晚，晚上
    date_all_afternoon = re.findall(r"(下午\d{1,2}[点:])", line)
    date_all_nightfall = re.findall(r"(傍晚\d{1,2}[点:])", line)
    date_all_night = re.findall(r"(晚上\d{1,2}[点:])", line)
    for data in date_all_afternoon + date_all_nightfall + date_all_night:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(下午)', '', data)
            hour = re.sub(r'(傍晚)', '', hour)
            hour = re.sub(r'(晚上)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day
            # 确定小时 和 分钟
            hour = int(hour)
            if hour <= 12:
                hour += 12
            return line, year, month, day, hour

    # 数字 中午
    date_all_1 = re.findall(r"(中午[01]?[123][点:])", line)
    date_all_2 = re.findall(r"(正午[01]?[123][点:])", line)
    for data in date_all_1 + date_all_2:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(中午)', '', data)
            hour = re.sub(r'(正午)', '', hour)
            hour = re.sub(r'([点:])', '', hour)
            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 确定小时 和 分钟
            hour = int(hour)
            return line, year, month, day, hour

    # 数字 凌晨
    date_all_early = re.findall(r"(凌晨[0]?[123456][点:])", line)
    for data in date_all_early:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            hour = re.sub(r'(凌晨)', '', data)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 确定小时 和 分钟
            hour = int(hour)
            return line, year, month, day, hour

    # 数字
    date_all = re.findall(r"(\d{1,2}[点:])", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'([点:])', '', data)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 确定小时 和 分钟
            hour = int(hour)
            return line, year, month, day, hour

    # 中文 上午
    date_all_morning_1 = re.findall(r"(上午[二]?[十]?[零两一二三四五六七八九十][点:])", line)
    date_all_morning_2 = re.findall(r"(早上[二]?[十]?[零两一二三四五六七八九十][点:])", line)
    for data in date_all_morning_1 + date_all_morning_2:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            hour = re.sub(r'(上午)', '', data)
            hour = re.sub(r'(早上)', '', hour)
            hour = re.sub(r'(分)', '', hour)
            hour = re.sub(r'(钟)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    hour = 25
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
            return line, year, month, day, hour

    # 中文 下午
    date_all_afternoon = re.findall(r"(下午[二]?[十]?[两一二三四五六七八九十][点:])", line)
    date_all_nightfall = re.findall(r"(傍晚[二]?[十]?[两一二三四五六七八九十][点:])", line)
    date_all_night = re.findall(r"(晚上[二]?[十]?[两一二三四五六七八九十][点:])", line)
    for data in date_all_afternoon + date_all_nightfall + date_all_night:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(下午)', '', data)
            hour = re.sub(r'(傍晚)', '', hour)
            hour = re.sub(r'(晚上)', '', hour)
            hour = re.sub(r'(分)', '', hour)
            hour = re.sub(r'(钟)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    hour = 25
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
            if hour <= 12:
                hour += 12
            return line, year, month, day, hour

    # 中文 中午
    date_all_1 = re.findall(r"(中午十[一二三][点:])", line)
    date_all_2 = re.findall(r"(正午十[一二三][点:])", line)
    for data in date_all_1 + date_all_2:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(中午)', '', data)
            hour = re.sub(r'(正午)', '', hour)
            hour = re.sub(r'(分)', '', hour)
            hour = re.sub(r'(钟)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    hour = 25
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
            return line, year, month, day, hour

    # 中文 中午
    date_all_1 = re.findall(r"(中午一[点:])", line)
    date_all_2 = re.findall(r"(正午一[点:])", line)
    for data in date_all_1 + date_all_2:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            hour = 13
            return line, year, month, day, hour

    # 中文 凌晨
    date_all_early = re.findall(r"(凌晨[二]?[十]?[一两二三四五六七八九十][点:])", line)
    for data in date_all_early:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(凌晨)', '', data)
            hour = re.sub(r'(分)', '', hour)
            hour = re.sub(r'(钟)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    hour = 25
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
            return line, year, month, day, hour

    # 数字
    date_all = re.findall(r"([二]?[十]?[两一二三四五六七八九十][点:])", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            hour = re.sub(r'(分)', '', data)
            hour = re.sub(r'(钟)', '', hour)
            hour = re.sub(r'([点:])', '', hour)

            # 确定年月日
            line, year, month, day = day_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month, day = day_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            day = now.day if day is None else day

            # 解析小时
            if len(hour) > 1:
                if '十' not in hour:
                    hour = 25
                else:
                    pre_hour = hour[0:hour.index('十')]
                    post_hour = hour[hour.index('十') + 1:]
                    # 10
                    if len(pre_hour) == 0 and len(post_hour) == 0:
                        hour = 10
                    # 11 ~ 19
                    elif len(pre_hour) == 0 and len(post_hour) != 0:
                        hour = 10 + int(time_map[post_hour])
                    # 20, 30
                    elif len(pre_hour) != 0 and len(post_hour) == 0:
                        hour = int(time_map[pre_hour]) * 10
                    else:
                        # 21 ~ 29
                        hour = int(time_map[pre_hour]) * 10 + int(time_map[post_hour])
            else:
                hour = int(time_map[hour])
            return line, year, month, day, hour

    if len(line) >= 16 and line[-1] in connect_char:
        holder = line[-16:-1]
        if holder in placeholders_list.keys():
            _, time_list = placeholders_list[holder]
            if len(time_list) == 1 and time_list[0] is not None:
                return line[:-1] + '到', time_list[0].year, time_list[0].month, time_list[0].day, time_list[0].hour

    return line, None, None, None, None



import sys
from os.path import dirname, abspath

path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import torch
import os
import numpy as np
import transformers
from Text2JSON.train import utils
from Text2JSON.model.base_model import BaseModel
from torch import nn


class HydraColumnTorch(BaseModel):
    def __init__(self, config):
        # 配置文件config
        self.config = config
        # 根据config设置，加载对应的HydraNet
        self.column_model = HydraColumnNet(config)
        # 定义优化器，学习率优化器
        self.optimizer, self.scheduler = None, None

    def init_optimizer(self):
        # 衰退设定
        no_decay = ["bias", "LayerNorm.weight"]
        # 返回model的所有参数的(name, tensor)的键值对, 并且设置其值
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self.column_model.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": float(self.config["decay"]),
            },
            {"params": [p for n, p in self.column_model.named_parameters() if any(nd in n for nd in no_decay)],
             "weight_decay": 0.0},
        ]
        # AdamW是实现了权重衰减的优化器
        self.optimizer = transformers.AdamW(optimizer_grouped_parameters,
                                            lr=float(self.config["column_learning_rate"]))
        # 设置耐心系数
        self.scheduler = transformers.get_cosine_schedule_with_warmup(
            self.optimizer,
            num_warmup_steps=int(self.config["column_num_warmup_steps"]),
            num_training_steps=int(self.config["column_num_train_steps"]))

        # 梯度清零
        self.optimizer.zero_grad()

    def train_on_batch(self, batch):
        self.column_model.train()
        # 加载数据到cuda
        if self.config['has_cuda']:
            for k, v in batch.items():
                batch[k] = v.to(torch.device("cuda:0"))

        # 默认调用model.forward函数
        batch_loss = torch.mean(self.column_model(**batch)["loss"])  # 计算loss

        batch_loss.backward()  # loss反向传播
        torch.nn.utils.clip_grad_norm_(self.column_model.parameters(), 1.0)  # 用L2范数作为正则
        self.optimizer.step()  # 优化
        self.scheduler.step()  # 更新学习率
        self.optimizer.zero_grad()  # 梯度清零

        if self.config['has_cuda']:
            # 返回一定需要detach, 否则会累积梯度
            return batch_loss.cpu().detach().numpy()
        else:
            return batch_loss.detach().numpy()

    # 模型预测
    def model_inference(self, model_inputs):
        # 不启用 BatchNormalization 和 Dropout
        self.column_model.eval()
        model_outputs = {}

        # batch_size = 512  # 设置步长， 这一步有可能出现内存溢出，内存太大了
        batch_size = int(self.config['column_evaluator_batch_size'])
        # 在(0~shape[0])之间遍历，步长为batch_size
        '''
            一个batch的输入格式
            input_tensor={
                            'input_ids': [],
                            'input_mask': [],
                            'segment_ids': []
                            }
        '''
        for start_idx in range(0, model_inputs["input_ids"].shape[0], batch_size):
            if self.config['has_cuda']:
                input_tensor = {
                    # 将batch中的数据分组，以batch_size为单位，分组
                    # 只取输入编码特征，不取其他特征
                    k: torch.from_numpy(model_inputs[k][start_idx:start_idx + batch_size]).to(torch.device("cuda:0"))
                    for k in ["input_ids", "input_mask", "segment_ids"]
                }
            else:
                input_tensor = {
                    # 将batch中的数据分组，以batch_size为单位，分组
                    # 只取输入编码特征，不取其他特征
                    k: torch.from_numpy(model_inputs[k][start_idx:start_idx + batch_size])
                    for k in ["input_ids", "input_mask", "segment_ids"]
                }
            '''
                一个batch输出格式
                model_output ={
                                'column_func': [[1,1,1], [1,1,1], ..., []]
                                'sel': [[],[],...,[]]
                                'sel_num':[1, 1, 1, ..., 1]
                                'agg': [[], [],..., []] 
                                'where_num': [1, 1, 1, ..., 1]
                                'op': [[], [],..., []]
                                'value_num':[1, 1, 1, 1]
                                'value_start': [[], [], ..., []]
                                'value_end': [[], [], ..., []]
                                'relationship': [1, 1, 1, 1, 1]
                                'loss': [num1, num2, ..., ]
                                }
            '''
            # 因为只做预测，所以不更新梯度
            with torch.no_grad():
                # 数据输入模型
                model_output = self.column_model(**input_tensor)

            # 将输出按关键字key进行分组
            for k, out_tensor in model_output.items():
                if out_tensor is None:
                    continue
                if k not in model_outputs:
                    model_outputs[k] = []
                if self.config['has_cuda']:
                    # 将数据转移到CPU，否则就会导致GPU内存溢出
                    model_outputs[k].append(out_tensor.cpu().detach().numpy())
                else:
                    model_outputs[k].append(out_tensor.detach().numpy())

        # 将多个batch结果合在一起
        for k in model_outputs:
            model_outputs[k] = np.concatenate(model_outputs[k], 0)

        '''
            concatenate前    [[[],[],...,[]], [[],[],...,[]]]
                                batch1          batch2
            concatenate后    [[],[],[],[],[],............[]]
            
            多个batch的最终输出
            model_outputs ={
                            'column_func': [[1,1,1], [1,1,1], ..., []]
                            'seq': [1, 1, 1, 0]
                            'sel': [[],[],...,[]]
                            'sel_num':[1, 1, 1, ..., 1]
                            'agg': [[], [],..., []] 
                            'where_num': [1, 1, 1, ..., 1]
                            'op': [[], [],..., []]
                            'value_num':[1, 1, 1, 1]
                            'value_start': [[], [], ..., []]
                            'value_end': [[], [], ..., []]
                            'relationship': [1, 1, 1, 1, 1]
                            'loss': [num1, num2, ..., ]
                            }
        '''
        return model_outputs

    def save(self, model_path, name):
        save_path = os.path.join(model_path, "model_{0}.pt".format(name))
        torch.save(self.column_model.state_dict(), save_path)
        print("Column Model saved in path: %s" % save_path)

    def load(self, model_path, name):
        pt_path = os.path.join(model_path, "model_{0}.pt".format(name))
        self.column_model.load_state_dict(torch.load(pt_path))
        print("Column Model loaded from {0}".format(pt_path))

    def to_cpu(self):
        self.column_model.cpu()

    def to_gpu(self):
        self.column_model.to(torch.device("cuda:0"))


class HydraColumnNet(nn.Module):
    def __init__(self, config):
        super(HydraColumnNet, self).__init__()
        self.config = config

        # 加载transformer
        self.base_model = utils.create_base_model(config)

        drop_rate = float(config["drop_rate"]) if "drop_rate" in config else 0.0
        self.dropout = nn.Dropout(drop_rate)

        bert_hid_size = self.base_model.config.hidden_size  # 获取隐变量维度

        # 输入维度是bert_hid_size, 输出是3维, Loss用Sigmod+Binary CrossEntropyLoss函数计算
        # SQL1: 第一个是select相似度     p(c∈S|Q)  sel 中所选择的列
        # SQL1: 第二个是where相似度      p(c∈W|Q)  where 中所选择的列
        # SQL1: 第三个是整体相似度        p(c∈R|Q)  SQL 语句的整体相似度
        self.sql_first_select_sim = nn.Linear(bert_hid_size, 2)
        self.sql_first_where_sim = nn.Linear(bert_hid_size, 2)
        self.sql_first_all_sim = nn.Linear(bert_hid_size, 2)

        # SQL2: 第一个是select相似度     p(c∈S|Q)  sel 中所选择的列
        # SQL2: 第二个是where相似度      p(c∈W|Q)  where 中所选择的列
        # SQL2: 第三个是整体相似度        p(c∈R|Q)  SQL 语句的整体相似度
        # self.column_func_two = nn.Linear(bert_hid_size, 3)
        self.sql_sec_select_sim = nn.Linear(bert_hid_size, 2)
        self.sql_sec_where_sim = nn.Linear(bert_hid_size, 2)
        self.sql_sec_all_sim = nn.Linear(bert_hid_size, 2)

        # 计算输入的列操作符，loss=CrossEntropy
        self.agg = nn.Linear(bert_hid_size, int(config["agg_num"]))  # 输入维度是bert_hid_size, 输出是agg_num维
        # 计算输入的条件操作符，loss=CrossEntropy
        self.op = nn.Linear(bert_hid_size, int(config["op_num"]))  # 输入维度是bert_hid_size, 输出是op_num维

        # +1 是为了让0 作为没有值对应的value point， 第0维 不作训练
        # 用于计算输入在Where_Clause对应的值数目nv，p(nv|c,q)，loss=CrossEntropy
        self.value_num = nn.Linear(bert_hid_size, int(config["max_value_num"]) + 1)
        # 就上列对应值的 start_index, 一共可以对应max_value_num个值, Loss = CrossEntropy
        # [batch_size, token_num, hid_size] -> [batch_size, token_num, max_value_num]
        self.value_start = nn.Linear(bert_hid_size, int(config["max_value_num"]) + 1)
        # 就上列对应值的 end_index, 一共可以对应max_value_num个值, Loss = CrossEntropy
        # [batch_size, token_num, hid_size] -> [batch_size, token_num, max_value_num]
        self.value_end = nn.Linear(bert_hid_size, int(config["max_value_num"]) + 1)
        # 计算SQL语句之间的关系信息, loss = CrossEntropy
        self.relationship = nn.Linear(bert_hid_size, 2)

    def forward(self, input_ids, input_mask, segment_ids, seq=None, sel_num=None, sel=None, agg=None,
                where_num=None, where=None, op=None, value_num=None, value_start=None, value_end=None,
                relationship=None):
        """
            question: question token 单个
            table_id: table id 单个
            tokens: [[cls]type,column[SEP]question[SEP],...] 一组 column和question的联合表示，获取编码后的序列
            word_to_char_start: [start1, start2, start3, start4] 一组 记录问题token字符的开始位置
            word_to_subword: [[(start_index, end_index),...,],...,[]] 一组 词 到 子词 的映射关系,形式
            subword_to_word: [[0,1,1,2,2,2],...] 一组 子词 到 词 的映射关系
            input_ids: [[id1, id2, id3,..., pad, pad],...] 一组 将输入tokens, 通过词表转换成id
            input_mask: [[1,1,1,1],...] 一组 标识column+question位置编码,也就是整个输入所占的编码
            segment_ids: [[0,0,1,1],...] 一组 标识column和question位置
            columns: [name1, name2, name3, name4, name5] question有可能所涉及到的列
            seq: [0, 0, 0, 0, 0] 位置信息0代表第一条SQL语句，1代表第二条SQL语句
            sel_num: [1, 1, 1, 1, 1] select_clause 对应列的数目
            sel: [0,0,1,0,0] select_clause 所选择的列
            agg: [0, 0, 1, 0, 0]有可能所涉及到的列对应操作符
            where_num: [1, 1, 1, 1, 1] 每一个<>column<>question<> pair 对应的列数目, [len(conds)] * columns
            where: [0, 0, 1, 0, 0] 记录每一个列是否出现在where_clause中
            op: [0, 0, 2, 0, 0] 记录在where_clause中，每一个列对应的操作符
            value_num: [1, 2, 3, 0, 0] where子句中列对应值的数目
            value_start: [[0,...], [0,...], [start_pos1, start_pos2,...], ...] 记录在where_clause中，每一个列对应值在tokens的开始位置, 如果没有，记为-1
            value_end: [[0], [0], [end_pos1, end_pos2] , [0], [0]] 记录在where_clause中，每一个列对应值在tokens的结束位置, 如果没有，记为-1
            relationship: [0, 0, 0, 0, 0] 结构信息 0:代表单条SQL 1:代表两条条SQL
        """

        # 获取输出
        # bert_output 是最后一层编码层的特征向量 <cls>h,h,...,h<sep>q,q,...,q<sep>
        # pooled_output 是 cls
        bert_output, pooled_output = self.base_model(
            input_ids=input_ids,
            attention_mask=input_mask,
            token_type_ids=segment_ids)

        # dropout 放弃部分值
        bert_output = self.dropout(bert_output)
        pooled_output = self.dropout(pooled_output)

        # SQL 1
        sql_first_select_logit = self.sql_first_select_sim(pooled_output)
        sql_first_where_logit = self.sql_first_where_sim(pooled_output)
        sql_first_all_logit = self.sql_first_all_sim(pooled_output)

        # SQL 2
        sql_sec_select_logit = self.sql_sec_select_sim(pooled_output)
        sql_sec_where_logit = self.sql_sec_where_sim(pooled_output)
        sql_sec_all_logit = self.sql_sec_all_sim(pooled_output)

        # self.agg          :  [CLS] bert_hid_size -> config["agg_num"]
        # self.op           :  [CLS] bert_hid_size -> config["op_num"]
        # self.value_num    :  [CLS] bert_hid_size -> config["max_value_num"] + 1
        # self.value_start  :  [SEQ] bert_hid_size -> config["max_value_num"] + 1
        # self.value_end    :  [SEQ] bert_hid_size -> config["max_value_num"] + 1
        # self.relationship :  [CLS] bert_hid_size -> 2

        # public
        agg_logit = self.agg(pooled_output)
        op_logit = self.op(pooled_output)
        value_num_logit = self.value_num(pooled_output)  # 如果不行这里可以改一下
        value_start_logit = self.value_start(bert_output)
        value_end_logit = self.value_end(bert_output)
        relationship_logit = self.relationship(pooled_output)

        # 转换数据类型
        # 这里应该是用segment_ids, 因为start_index 和 end_index都不可能在column中，只能在question中
        value_span_mask = input_mask.to(dtype=bert_output.dtype)

        for index in range(0, int(self.config['max_value_num'])):
            value_start_logit[:, :, index + 1] = value_start_logit[:, :, index + 1] * value_span_mask - 1000000.0 * (
                    1 - value_span_mask)
            value_end_logit[:, :, index + 1] = value_end_logit[:, :, index + 1] * value_span_mask - 1000000.0 * (
                    1 - value_span_mask)

        loss = None
        # select is not None 对应train的情况
        # select is None 对应inference的情况
        if sel is not None:
            bceloss = nn.BCEWithLogitsLoss(reduction="none")  # 将sigmod函数和Binary CrossEntropyLoss合成一步
            cross_entropy = nn.CrossEntropyLoss(reduction="none")  # Cross Entropy Loss对应softmax函数的loss function

            select_rel = sel.float()
            where_rel = where.float()
            sql_rel = (sel.float() + where.float() - sel.float() * where.float())
            # SQL1 similarity loss
            loss = bceloss(sql_first_select_logit[:, 0], select_rel) * (1 - seq)
            loss += bceloss(sql_first_where_logit[:, 0], where_rel) * (1 - seq)
            loss += bceloss(sql_first_all_logit[:, 0], sql_rel) * (1 - seq)

            # SQL2 similarity loss
            loss += bceloss(sql_sec_select_logit[:, 0], select_rel) * seq
            loss += bceloss(sql_sec_where_logit[:, 0], where_rel) * seq
            loss += bceloss(sql_sec_all_logit[:, 0], sql_rel) * seq

            # SQL1 和 SQL2 可以共用的部分
            loss += cross_entropy(agg_logit, agg) * select_rel
            loss += cross_entropy(op_logit, op) * where_rel
            loss += cross_entropy(value_num_logit, value_num) * where_rel
            for index in range(0, int(self.config['max_value_num'])):
                loss += cross_entropy(value_start_logit[:, :, index + 1], value_start[:, index]) * where_rel
                loss += cross_entropy(value_end_logit[:, :, index + 1], value_end[:, index]) * where_rel
            loss += cross_entropy(relationship_logit, relationship) * sql_rel

        log_sigmoid = nn.LogSigmoid()
        return {"sql_first_select_sim": log_sigmoid(sql_first_select_logit),
                "sql_first_where_sim": log_sigmoid(sql_first_where_logit),
                "sql_first_all_sim": log_sigmoid(sql_first_all_logit),
                "sql_sec_select_sim": log_sigmoid(sql_sec_select_logit),
                "sql_sec_where_sim": log_sigmoid(sql_sec_where_logit),
                "sql_sec_all_sim": log_sigmoid(sql_sec_all_logit),
                "agg": agg_logit.log_softmax(1),
                "op": op_logit.log_softmax(1),
                "value_num": value_num_logit.log_softmax(1),
                "value_start": value_start_logit.log_softmax(1),
                "value_end": value_end_logit.log_softmax(1),
                "relationship": relationship_logit.log_softmax(1),
                "loss": loss}



#!/usr/bin/env python
from __future__ import print_function
from setuptools import setup, find_packages

setup(
    name="Text2JSON",
    version="0.1.4",
    author="Yucheng Zeng",
    author_email="yucheng-zeng@outlook.com",
    description="Text2JSON",
    long_description=open("README.md").read(),
    license="MIT",
    url="https://github.com/1173710105/Text2JSON",
    packages=find_packages(),
    install_requires=[
        "numpy",
        'transformers==2.5.1',
        'chinese2digits'
    ],
    python_requires='>=3',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
    ],
    zip_safe=False,
)



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import datetime
import re
from datetime import date, timedelta
from Text2JSON.entity_named_recog.month_entity_recognition import month_absolute_comfirm, month_relative_comfirm
from Text2JSON.entity_named_recog.entity_utils import *


def day_recon(line, placeholders_list: dict):
    line = day_relative_recognition(line, placeholders_list)
    line = day_absolute_recognition(line, placeholders_list)
    return line


def day_absolute_recognition(line, placeholders_list: dict):
    # 需要找到最后出现的位置, '....2013年...' 不识别, '...2014年'识别
    # 重复出现,"2014年...,2014年"只识别最后一个
    now = date.today()
    day_map = {'一': '1', '二': '2', '三': '3', '四': '4',
               '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}

    # 出日01-31日
    date_all = re.findall(r"(\d{1,2}日)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        day = int(data.replace('日', ''))

        # 确定年月
        handle_line, year, month = month_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month = month_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month
        try:
            # 解析不出来的日期需要跳过
            target_day = datetime.datetime(year, month, day)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_day is not None, '逻辑错误'
        time_str = '“' + date_to_str(target_day) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_day])
        line = handle_line + placeholder + sec_line

    # 处理 十 ~ 三十一 中文日期
    date_all = re.findall(r'([二三]?[十][一二三四五六七八九]?日)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 确定年月
        handle_line, year, month = month_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month = month_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month

        # 转成数字
        data_digit = data.replace('日', '')
        prefix = data_digit[0:data_digit.index('十')]
        postfix = data_digit[data_digit.index('十') + 1:]
        # 10
        if len(prefix) == 0 and len(postfix) == 0:
            day = 10
        # 11 ~ 19
        elif len(prefix) == 0 and len(postfix) != 0:
            day = 10 + int(day_map[postfix])
        # 20, 30
        elif len(prefix) != 0 and len(postfix) == 0:
            day = int(day_map[prefix]) * 10
        else:
            # 21 ~ 29
            day = int(day_map[prefix]) * 10 + int(day_map[postfix])
        try:
            # 解析不出来的日期需要跳过
            target_day = datetime.datetime(year, month, day)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_day is not None, '逻辑错误'
        time_str = '“' + date_to_str(target_day) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_day])
        line = handle_line + placeholder + sec_line

    # 处理 一 ~ 九 中文日期
    date_all = re.findall(r'([一二三四五六七八九]日)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 确定年月
        handle_line, year, month = month_relative_comfirm(first_line, placeholders_list)
        if year is None:
            handle_line, year, month = month_absolute_comfirm(first_line, placeholders_list)
        year = now.year if year is None else year
        month = now.month if month is None else month

        # 转成数字
        day = int(day_map[data.replace('日', '')])

        try:
            # 解析不出来的日期需要跳过
            target_day = datetime.datetime(year, month, day)
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        assert target_day is not None, '逻辑错误'
        time_str = '“' + date_to_str(target_day) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [target_day])
        line = handle_line + placeholder + sec_line

    return line


def day_relative_recognition(line, placeholders_list: dict):
    cur_day = ['今天', '这天']
    next_day = ['明天', '下一天', '一天后', '后一天']
    last_day = ['昨天', '上一天', '一天前', '前一天']
    day_map = {'一': '1', '二': '2', '三': '3', '四': '4',
               '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '两': '2', '十': '10'}
    for day in cur_day:
        if day in line:
            index = line.index(day)
            first_line = line[0: index]
            sec_line = line[index + len(day):]

            # now
            now = date.today()

            time_str = '“' + date_to_str(now) + "”"
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [now])
            line = first_line + placeholder + sec_line

    for day in next_day:
        if day in line:
            index = line.index(day)
            first_line = line[0: index]
            sec_line = line[index + len(day):]

            # now
            now = date.today()

            next_day_date = now + timedelta(days=1)

            # to str and insert into line
            time_str = '“' + date_to_str(next_day_date) + "”"
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [next_day_date])
            line = first_line + placeholder + sec_line

    for day in last_day:
        if day in line:
            index = line.index(day)
            first_line = line[0: index]
            sec_line = line[index + len(day):]

            now = date.today()
            # 前一天
            next_n_date = now + timedelta(days=-1)

            # to str and insert into line
            time_str = '“' + date_to_str(next_n_date) + '”'
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [next_n_date])
            line = first_line + placeholder + sec_line

    date_all = re.findall(r'(\d{1,2}[日天]后)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        day_delta = int(re.sub(r'[日天]后', '', data))
        now = date.today()
        # 前一天
        next_n_date = now + timedelta(days=int(day_delta))

        # to str and insert into line
        time_str = '“' + date_to_str(next_n_date) + '”'
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [next_n_date])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r'([二三四五六七八九]?[十]?[一两二三四五六七八九十][日天]后)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        day_delta = re.sub(r'[日天]后', '', data)
        now = date.today()

        try:
            # 解析分钟
            if len(day_delta) > 1:
                if '十' not in day_delta:
                    # 错误处理
                    raise ValueError
                else:
                    pre_day = day_delta[0:day_delta.index('十')]
                    post_minute = day_delta[day_delta.index('十') + 1:]
                    # 10
                    if len(pre_day) == 0 and len(post_minute) == 0:
                        day = 10
                    # 11 ~ 19
                    elif len(pre_day) == 0 and len(post_minute) != 0:
                        day = 10 + int(day_map[post_minute])
                    # 20, 30
                    elif len(pre_day) != 0 and len(post_minute) == 0:
                        day = int(day_map[pre_day]) * 10
                    else:
                        # 21 ~ 24
                        day = int(day_map[pre_day]) * 10 + int(day_map[post_minute])
            else:
                day = int(day_map[day_delta])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue
        # 前一天
        next_n_date = now + timedelta(days=day)

        # to str and insert into line
        time_str = '“' + date_to_str(next_n_date) + '”'
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [next_n_date])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r'(\d{1,2}[日天]前)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        day_delta = int(re.sub(r'[日天]前', '', data))
        now = date.today()
        # 前一天
        last_n_date = now - timedelta(days=int(day_delta))

        # to str and insert into line
        time_str = '“' + date_to_str(last_n_date) + '”'
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [last_n_date])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r'([二三四五六七八九]?[十]?[一两二三四五六七八九十][日天]前)', line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        day_delta = re.sub(r'[日天]前', '', data)
        now = date.today()

        try:
            # 解析分钟
            if len(day_delta) > 1:
                if '十' not in day_delta:
                    # 错误处理
                    raise ValueError
                else:
                    pre_day = day_delta[0:day_delta.index('十')]
                    post_minute = day_delta[day_delta.index('十') + 1:]
                    # 10
                    if len(pre_day) == 0 and len(post_minute) == 0:
                        day = 10
                    # 11 ~ 19
                    elif len(pre_day) == 0 and len(post_minute) != 0:
                        day = 10 + int(day_map[post_minute])
                    # 20, 30
                    elif len(pre_day) != 0 and len(post_minute) == 0:
                        day = int(day_map[pre_day]) * 10
                    else:
                        # 21 ~ 24
                        day = int(day_map[pre_day]) * 10 + int(day_map[post_minute])
            else:
                day = int(day_map[day_delta])
        except ValueError:
            # 识别不出来就认为是错误，后续不再识别
            time_str = data
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [None])
            line = first_line + placeholder + sec_line
            continue

        last_n_date = now - timedelta(days=day)
        # to str and insert into line
        time_str = '“' + date_to_str(last_n_date) + '”'
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [last_n_date])
        line = first_line + placeholder + sec_line
    return line


def day_relative_comfirm(line, placeholders_list: dict):
    cur_day = ['今天', '这天']
    next_day = ['明天', '下一天', '一天后', '后一天', '下1天', '1天后', '后1天']
    last_day = ['昨天', '上一天', '一天前', '前一天', '上一天', '一天前', '前一天']
    day_map = {'一': '1', '二': '2', '三': '3', '四': '4',
               '五': '5', '六': '6', '七': '7', '八': '8', '九': '9', '两': '2'}
    # 连字符 = []
    now = date.today()
    for data in cur_day:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                line = line[:index]
                return line, now.year, now.month, now.day

    for data in next_day:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                line = line[:index]
                next_day_date = now + timedelta(days=1)
                return line, next_day_date.year, next_day_date.month, next_day_date.day

    for data in last_day:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                line = line[:index]
                last_day_date = now - timedelta(days=1)
                return line, last_day_date.year, last_day_date.month, last_day_date.day

    date_all = re.findall(r'(\d{1,2}[日天]后)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            day_delta = int(re.sub(r'[日天]后', '', data))
            now = date.today()
            next_n_date = now + timedelta(days=int(day_delta))
            return line, next_n_date.year, next_n_date.month, next_n_date.day

    date_all = re.findall(r'([二三四五六七八九]十[一二三四五六七八九]?[日天]后)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            day_delta = re.sub(r'[日天]后', '', data)
            now = date.today()

            prefix = day_delta[0:day_delta.index('十')]
            postfix = day_delta[day_delta.index('十') + 1:]
            # 10
            if len(prefix) == 0 and len(postfix) == 0:
                day = 10
            # 11 ~ 19
            elif len(prefix) == 0 and len(postfix) != 0:
                day = 10 + int(day_map[postfix])
            # 20, 30
            elif len(prefix) != 0 and len(postfix) == 0:
                day = int(day_map[prefix]) * 10
            else:
                # 21 ~ 29
                day = int(day_map[prefix]) * 10 + int(day_map[postfix])

            # 前一天
            next_n_date = now + timedelta(days=day)
            return line, next_n_date.year, next_n_date.month, next_n_date.day

    date_all = re.findall(r'([两一二三四五六七八九][日天]后)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            day_delta = re.sub(r'[日天]后', '', data)
            now = date.today()
            next_n_date = now + timedelta(days=int(day_map[day_delta]))
            return line, next_n_date.year, next_n_date.month, next_n_date.day

    date_all = re.findall(r'(\d{1,2}[日天]前)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            day_delta = int(re.sub(r'[日天]前', '', data))
            now = date.today()
            # 前一天
            last_n_date = now - timedelta(days=int(day_delta))
            return line, last_n_date.year, last_n_date.month, last_n_date.day

    date_all = re.findall(r'([二三四五六七八九]十[一二三四五六七八九]?[日天]前)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            day_delta = re.sub(r'[日天]前', '', data)
            now = date.today()

            prefix = day_delta[0:day_delta.index('十')]
            postfix = day_delta[day_delta.index('十') + 1:]
            # 10
            if len(prefix) == 0 and len(postfix) == 0:
                day = 10
            # 11 ~ 19
            elif len(prefix) == 0 and len(postfix) != 0:
                day = 10 + int(day_map[postfix])
            # 20, 30
            elif len(prefix) != 0 and len(postfix) == 0:
                day = int(day_map[prefix]) * 10
            else:
                # 21 ~ 29
                day = int(day_map[prefix]) * 10 + int(day_map[postfix])

            # 前一天
            next_n_date = now - timedelta(days=day)
            return line, next_n_date.year, next_n_date.month, next_n_date.day

    date_all = re.findall(r'([两一二三四五六七八九][日天]前)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            day_delta = re.sub(r'[日天]前', '', data)
            now = date.today()

            last_n_date = now - timedelta(days=int(day_map[day_delta]))
            return line, last_n_date.year, last_n_date.month, last_n_date.day

    return line, None, None, None


def day_absolute_comfirm(line, placeholders_list: dict):
    now = date.today()
    day_map = {'一': '1', '二': '2', '三': '3', '四': '4',
               '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}
    connect_char = ['-', '~', '到', '至']
    # 出日01-31日
    date_all = re.findall(r"(\d{1,2}日)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            # 确定年月日
            day = int(data.replace('日', ''))
            line, year, month = month_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month = month_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month
            return line, year, month, day

    # 处理 十 ~ 三十一 中文日期
    date_all = re.findall(r'([二三]?[十][一二三四五六七八九]?日)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]

            # 确定年月
            line, year, month = month_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month = month_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month

            # 转成数字
            data_digit = data.replace('日', '')
            prefix = data_digit[0:data_digit.index('十')]
            postfix = data_digit[data_digit.index('十') + 1:]
            # 10
            if len(prefix) == 0 and len(postfix) == 0:
                day = 10
            # 11 ~ 19
            elif len(prefix) == 0 and len(postfix) != 0:
                day = 10 + int(day_map[postfix])
            # 20, 30
            elif len(prefix) != 0 and len(postfix) == 0:
                day = int(day_map[prefix]) * 10
            else:
                # 21 ~ 29
                day = int(day_map[prefix]) * 10 + int(day_map[postfix])
            return line, year, month, day

    # 处理 一 ~ 九 中文日期
    date_all = re.findall(r'([一二三四五六七八九]日)', line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            line = line[:index]
            # 确定年月
            line, year, month = month_relative_comfirm(line, placeholders_list)
            if year is None:
                line, year, month = month_absolute_comfirm(line, placeholders_list)
            year = now.year if year is None else year
            month = now.month if month is None else month

            # 转成数字
            day = int(day_map[data.replace('日', '')])
            return line, year, month, day

    if len(line) >= 16 and line[-1] in connect_char:
        holder = line[-16:-1]
        if holder in placeholders_list.keys():
            _, time_list = placeholders_list[holder]
            if len(time_list) == 1 and time_list[0] is not None:
                return line[:-1] + '到', time_list[0].year, time_list[0].month, time_list[0].day

    return line, None, None, None



import sys
from os.path import dirname, abspath
path = dirname(dirname(dirname(abspath(__file__))))
sys.path.append(path)
import re
import datetime
from datetime import date
from Text2JSON.entity_named_recog.entity_utils import *


def year_recon(line, placeholders_list: dict):
    line = year_relative_recognition(line, placeholders_list)
    line = year_absolute_recognition(line, placeholders_list)
    return line


def year_absolute_recognition(line, placeholders_list: dict):
    year_map = {'零': '0', '一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}
    date_all = re.findall(r"(\d{4}年)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 确定年的开始时间和结束时间
        year = int(data.replace('年', ''))
        year_start = datetime.datetime(year, 1, 1)
        year_end = datetime.datetime(year, 12, 31)

        time_str = '“' + date_to_str(year_start) + "”" + "到" + '“' + date_to_str(year_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [year_start, year_end])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r"(\d{2}年)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        year = int('20' + data.replace('年', ''))
        year_start = datetime.datetime(year, 1, 1)
        year_end = datetime.datetime(year, 12, 31)

        time_str = '“' + date_to_str(year_start) + "”" + "到" + '“' + date_to_str(year_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [year_start, year_end])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r"([零一二三四五六七八九]{4}年)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 转成中文数字
        data = data.replace('年', '')
        year_str = ''
        for d_char in data:
            year_str += year_map[d_char]
        year = int(year_str)
        year_start = datetime.datetime(year, 1, 1)
        year_end = datetime.datetime(year, 12, 31)

        time_str = '“' + date_to_str(year_start) + "”" + "到" + '“' + date_to_str(year_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [year_start, year_end])
        line = first_line + placeholder + sec_line

    date_all = re.findall(r"([零一二三四五六七八九]{2}年)", line)
    for data in date_all:
        index = line.index(data)
        first_line = line[0: index]
        sec_line = line[index + len(data):]

        # 转成中文数字
        data = data.replace('年', '')
        year_str = ''
        for d_char in data:
            year_str += year_map[d_char]
        year = int('20' + year_str)
        year_start = datetime.datetime(year, 1, 1)
        year_end = datetime.datetime(year, 12, 31)

        time_str = '“' + date_to_str(year_start) + "”" + "到" + '“' + date_to_str(year_end) + "”"
        placeholder = random_str()
        placeholders_list[placeholder] = (time_str, [year_start, year_end])
        line = first_line + placeholder + sec_line

    return line


def year_relative_recognition(line, placeholders_list: dict):
    now_year = ['最近一年', '近一年', '最近1年', '近1年', '近年']
    cur_year = ['本年度', '本年', '今年', '这年']
    next_year = ['下1年', '下一年', '下年', '明年', '一年后']
    last_year = ['上1年', '上一年', '上年', '去年', '一年前']
    flag = True
    # 保证最长匹配优先, 因为存在共同后缀
    while flag:
        flag = False
        for year in now_year:
            if year in line:
                index = line.index(year)
                first_line = line[0: index]
                sec_line = line[index + len(year):]

                # now
                now = date.today()
                # 去年的同一时间
                try:
                    last_year_now = datetime.datetime(now.year - 1, now.month, now.day)
                except ValueError:
                    # 处理2月29日这种情况
                    last_year_now = datetime.datetime(now.year - 1, now.month, now.day - 1)

                # to str and insert into line
                time_str = '“' + date_to_str(last_year_now) + "”" + "到" + '“' + date_to_str(now) + "”"
                placeholder = random_str()
                placeholders_list[placeholder] = (time_str, [last_year_now, now])
                line = first_line + placeholder + sec_line
                flag = True
                break

    for year in cur_year:
        if year in line:
            index = line.index(year)
            first_line = line[0: index]
            sec_line = line[index + len(year):]

            # now
            now = date.today()
            # last year
            cur_year_start = datetime.datetime(now.year, 1, 1)
            cur_year_end = datetime.datetime(now.year, 12, 31)

            # to str and insert into line
            time_str = '“' + date_to_str(cur_year_start) + "”" + "到" + '“' + date_to_str(cur_year_end) + "”"
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [cur_year_start, cur_year_end])
            line = first_line + placeholder + sec_line

    for year in next_year:
        if year in line:
            index = line.index(year)
            first_line = line[0: index]
            sec_line = line[index + len(year):]

            now = date.today()
            # 明年第一天喝最后一天
            next_year_start = datetime.datetime(now.year + 1, 1, 1)
            next_year_end = datetime.datetime(now.year + 1, 12, 31)

            # to str and insert into line
            time_str = '“' + date_to_str(next_year_start) + '”' + '到' + '“' + date_to_str(next_year_end) + '”'
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [next_year_start, next_year_end])
            line = first_line + placeholder + sec_line

    for year in last_year:
        if year in line:
            index = line.index(year)
            first_line = line[0: index]
            sec_line = line[index + len(year):]

            now = date.today()
            # 明年第一天喝最后一天
            last_year_start = datetime.datetime(now.year - 1, 1, 1)
            last_year_end = datetime.datetime(now.year - 1, 12, 31)

            # to str and insert into line
            time_str = '“' + date_to_str(last_year_start) + '”' + '到' + '“' + date_to_str(last_year_end) + '”'
            placeholder = random_str()
            placeholders_list[placeholder] = (time_str, [last_year_start, last_year_end])
            line = first_line + placeholder + sec_line

    return line


def year_absolute_comfirm(line, placeholders_list: dict):
    # 需要找到最后出现的位置, '....2013年...' 不识别, '...2014年'识别
    # 重复出现,"2014年...,2014年"只识别最后一个
    year = None
    yaer_map = {'零': '0', '一': '1', '二': '2', '三': '3', '四': '4',
                '五': '5', '六': '6', '七': '7', '八': '8', '九': '9'}
    connect_char = ['-', '~', '到', '至']

    date_all = re.findall(r"(\d{4}年)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            year = int(data.replace('年', ''))
            line = line[:index]
            return line, year

    date_all = re.findall(r"(\d{2}年)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            year = int('20' + data.replace('年', ''))
            line = line[:index]
            return line, year

    date_all = re.findall(r"([零一二三四五六七八九]{4}年)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            # 转成中文数字
            data = data.replace('年', '')
            year_str = ''
            for d_char in data:
                year_str += yaer_map[d_char]
            year = int(year_str)
            line = line[:index]
            return line, year

    date_all = re.findall(r"([零一二三四五六七八九]{2}年)", line)
    for data in date_all:
        index = line.rindex(data)
        if index + len(data) == len(line):
            # 转成中文数字
            data = data.replace('年', '')
            year_str = ''
            for d_char in data:
                year_str += yaer_map[d_char]
            year = int('20' + year_str)
            line = line[:index]
            return line, year

    if len(line) >= 16 and line[-1] in connect_char:
        holder = line[-16:-1]
        if holder in placeholders_list.keys():
            _, time_list = placeholders_list[holder]
            if len(time_list) == 1 and time_list[0] is not None:
                return line[:-1] + '到', time_list[0].year

    return line, year


def year_relative_comfirm(line, placeholders_list: dict):
    # 只识别最后一个词
    cur_year = ['今年', '本年', '这年', '近年', '最近一年']
    next_year = ['下一年', '下年', '明年', '下1年']
    last_year = ['上一年', '上年', '去年', '上1年']
    year = None
    now = date.today()
    for data in cur_year:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                year = now.year
                line = line[:index]
                return line, year
    for data in next_year:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                year = now.year + 1
                line = line[:index]
                return line, year

    for data in last_year:
        if data in line:
            index = line.rindex(data)
            if index + len(data) == len(line):
                year = now.year - 1
                line = line[:index]
                return line, year
    return line, year
